"""
Calculation of the transient climate response (TCR)
"""
import logging

import scmdata

from .base import _CalculatorTCRTCREBase
from .utils import _add_index_level, _time_mean

LOGGER = logging.getLogger(__name__)


class CalculatorTCRE(_CalculatorTCRTCREBase):
    """
    Calculator of the transient climate response to emissions (TCRE)
    """

    @classmethod
    def _can_calculate_metric(cls, metric):
        return metric == "Transient Climate Response to Emissions"

    @classmethod
    def _calculate(
        cls, assessed_ranges, res_calc, norm_period, evaluation_period, unit
    ):
        res_calc_normed = cls._get_normed_res_calc(
            assessed_ranges, res_calc, norm_period, evaluation_period
        )

        common_filters = dict(
            scenario="1pctCO2",
            region="World",
        )
        gsat_var = "Surface Air Temperature Change"
        cumulative_emms_var = "Cumulative Emissions|CO2"
        temperatures = res_calc_normed.filter(variable=gsat_var, **common_filters)
        cumulative_emissions = res_calc_normed.filter(
            variable=cumulative_emms_var, **common_filters
        )

        tcre = []
        for temperatures_cm in temperatures.groupby("climate_model"):
            climate_model = temperatures_cm.get_unique_meta("climate_model", True)
            cumulative_emissions_cm = cumulative_emissions.filter(
                climate_model=climate_model, log_if_empty=False
            )
            if cumulative_emissions_cm.empty:
                LOGGER.warning(
                    "No {} for `{}`".format(cumulative_emms_var, climate_model)
                )
                continue

            tcre_cm = temperatures_cm.divide(
                cumulative_emissions_cm, op_cols={"variable": "TCRE"}
            )
            tcre.append(tcre_cm)

        tcre = scmdata.run_append(tcre)

        out = _time_mean(tcre.filter(year=range(1920, 1921)).convert_unit(unit))

        out = _add_index_level(
            out.reset_index("variable", drop=True),
            ",".join([gsat_var, cumulative_emms_var]),
            "variable",
        )

        return out



"""
Utility functions for calculating metrics
"""
import pandas as pd


def _add_index_level(inp, level_val, level_name):
    return pd.concat({level_val: inp}, names=[level_name])


def _time_mean(inp, variable=None):
    """
    Take time mean of data

    Parameters
    ----------
    inp : :obj:`scmdata.ScmRun`
        Data of which to take the time mean

    variable : str
        Value of the variable column in the output. If ``None``, the values
        are not altered from those in ``inp``.

    Returns
    -------
    :obj:`pd.Series`
        Time mean of the data
    """
    out = inp.timeseries().mean(axis="columns")
    out.name = "value"
    out = _add_index_level(out, inp["year"].min(), "evaluation_period_start_year")
    out = _add_index_level(out, inp["year"].max(), "evaluation_period_end_year")

    if variable is not None:
        out = _add_index_level(
            out.reset_index("variable", drop=True), variable, "variable"
        )

    return out


def _regression_slope(inp, unit):
    """
    Get regression slope of data

    Parameters
    ----------
    inp : :obj:`scmdata.ScmRun`
        Data of which to take the time mean

    unit : str
        Unit of the regression slope

    Returns
    -------
    :obj:`pd.Series`
        Regression slopes of data
    """
    out = inp.linear_regression_gradient(unit=unit)
    out = out.rename({"gradient": "value"}, axis="columns")

    new_index = list(set(out.columns) - {"value"})
    out = out.set_index(new_index)["value"]
    out = _add_index_level(out, inp["year"].min(), "evaluation_period_start_year")
    out = _add_index_level(out, inp["year"].max(), "evaluation_period_end_year")

    return out



"""
Calculation of airborne fraction (of CO2)
"""
import logging

import numpy as np
import scmdata

from .base import Calculator
from .utils import _add_index_level, _time_mean

LOGGER = logging.getLogger(__name__)


class CalculatorAirborneFraction18501920(Calculator):
    """
    Calculator of the airborne fraction from 1850 to 1920
    """

    _start_year = 1850
    _end_year = 1920

    @classmethod
    def _can_calculate_metric(cls, metric):
        return metric == "Airborne Fraction|CO2 World 1pctCO2 1850-{}".format(
            cls._end_year
        )

    @classmethod
    def _calculate(
        cls, assessed_ranges, res_calc, norm_period, evaluation_period, unit
    ):
        cls._check_norm_period_evaluation_period(
            norm_period=norm_period,
            expected_norm_period=cls._start_year,
            evaluation_period=evaluation_period,
            expected_evaluation_period=cls._end_year,
        )

        common_filters = dict(
            scenario="1pctCO2",
            region="World",
            year=range(cls._start_year, cls._end_year + 1),
        )

        atmos_var = "Carbon Pool|Atmosphere"
        ocean_flux_var = "Net Atmosphere to Ocean Flux|CO2"
        land_flux_var = "Net Atmosphere to Land Flux|CO2"
        ocean_and_land_flux_var = "Net Atmosphere to Ocean and Land Flux|CO2"

        afs = []
        for res_calc_cm in res_calc.groupby("climate_model"):
            climate_model = res_calc_cm.get_unique_meta("climate_model", True)
            variables = res_calc_cm.get_unique_meta("variable")

            atmos_pool = res_calc_cm.filter(variable=atmos_var, **common_filters)
            if ocean_flux_var in variables and land_flux_var in variables:
                LOGGER.info(
                    "Using separate land and ocean fluxes for `{}`".format(
                        climate_model
                    )
                )
                ocean_fluxes = res_calc_cm.filter(
                    variable=ocean_flux_var, **common_filters
                )
                land_fluxes = res_calc_cm.filter(
                    variable=land_flux_var, **common_filters
                )

                assert not np.isnan(ocean_fluxes.values).any()
                assert not np.isnan(land_fluxes.values).any()

                ocean_and_land_flux = ocean_fluxes.add(
                    land_fluxes, op_cols={"variable": ocean_and_land_flux_var}
                )

            elif ocean_and_land_flux_var in variables:
                LOGGER.info(
                    "Using combined land and ocean fluxes for `{}`".format(
                        climate_model
                    )
                )
                ocean_and_land_flux = res_calc_cm.filter(
                    variable=ocean_and_land_flux_var, **common_filters
                )

                assert not np.isnan(ocean_and_land_flux.values).any()
            else:
                LOGGER.warning("No land or ocean fluxes for `{}`".format(climate_model))
                continue

            atmos_pool = assessed_ranges._get_normed_res_calc(atmos_pool, norm_period)

            ocean_and_land_pool = ocean_and_land_flux.integrate(
                "Carbon Pool|Ocean and Land"
            )
            ocean_and_land_pool = assessed_ranges._get_normed_res_calc(
                ocean_and_land_pool, norm_period
            )

            assert not np.isnan(atmos_pool.values).all()
            assert not np.isnan(ocean_and_land_pool.values).all()

            af = atmos_pool.divide(
                atmos_pool.add(
                    ocean_and_land_pool,
                    op_cols={"variable": "Atmos + Ocean + Land pool"},
                ),
                op_cols={"variable": "Airborne Fraction|CO2"},
            )
            afs.append(af)

        afs = scmdata.run_append(afs)
        out = _time_mean(
            afs.filter(year=range(cls._end_year, cls._end_year + 1)).convert_unit(unit)
        )

        out = _add_index_level(
            out.reset_index("variable", drop=True),
            ",".join([atmos_var, ocean_flux_var, land_flux_var]),
            "variable",
        )

        to_return = out.reset_index()
        to_return_idx = list(set(to_return.columns.tolist()) - {"value"})
        to_return = to_return.set_index(to_return_idx)["value"]

        return to_return


class CalculatorAirborneFraction18501990(CalculatorAirborneFraction18501920):
    """
    Calculator of the airborne fraction from 1850 to 1990
    """

    _start_year = 1850
    _end_year = 1990



"""
Validation of RCMIP submissions
"""

from logging import getLogger

import pandas as pd
import pint.errors
from scmdata.run import run_append

from ..errors import ProtocolConsistencyError
from .utils import load_submission_template_definitions

LOGGER = getLogger(__name__)


def validate_submission_bundle(timeseries, metadata, comments=None, protocol=None):
    """
    Validate that an RCMIP submission bundle complies with the required formats

    Parameters
    ----------
    timeseries : :obj:`scmdata.ScmRun`
        Timeseries to validate

    metadata : :obj:`pd.DataFrame`
        Model metadata

    comments : :obj:`pd.DataFrame`
        Submission comments

    protocol : str
        Data file containing the RCMIP protocol against which to validate the
        timeseries. If ``None``, the submission template will be loaded from
        ``pyrcmip/data/rcmip-data-submission-template-v4-0-0.xlsx``.

    Returns
    -------
    (:obj:`scmdata.ScmRun`, :obj:`pd.DataFrame`)
        Validated timeseries and model metadata

    Raises
    ------
    ProtocolConsistencyError
        The submission bundle is not consistent with the RCMIP protocol

    ValueError
        A value for ``climate_model`` is found in ``timeseries`` or ``model_reported``
        but isn't found in the ``climate_model`` column of ``metadata``.
    """
    raise_error = False

    try:
        metadata_consistent = validate_submission_model_meta(metadata)
        LOGGER.info("valid metadata")
    except ProtocolConsistencyError:
        LOGGER.exception("problems with metadata in submission")
        raise_error = True

    # Model metrics are not requested in the latest RCMIP round, but keeping it
    # around in case we want to re-enable it in the future.
    # try:
    #     model_reported_consistent = validate_submission_model_reported_metrics(
    #         model_reported
    #     )
    #     LOGGER.info("valid model reported csv")
    # except ProtocolConsistencyError:
    #     LOGGER.exception("problems with model reported csv in submission")
    #     raise_error = True

    if comments is not None:
        try:
            comments_consistent = validate_submission_model_comments(comments)
            LOGGER.info("valid comments")
        except Exception as e:
            LOGGER.exception(e)
            LOGGER.exception("problems with comments in submission")
            raise_error = True
    else:
        comments_consistent = None

    try:
        timeseries_consistent = validate_submission(timeseries, protocol=protocol)
        LOGGER.info("valid timeseries")
    except ProtocolConsistencyError:
        LOGGER.exception("problems with timeseries in submission")
        raise_error = True

    if raise_error:
        raise ProtocolConsistencyError(
            "submission bundle is inconsistent with the RCMIP protocol"
        )

    climate_model_consistency = True

    metadata_climate_models = metadata_consistent["climate_model"].unique().tolist()
    timeseries_climate_models = timeseries_consistent.get_unique_meta("climate_model")
    # model_reported_climate_models = (
    #     model_reported_consistent["climate_model"].unique().tolist()
    # )
    if comments is not None:
        comments_climate_models = comments_consistent["climate_model"].unique().tolist()

    timeseries_climate_models_extras = set(timeseries_climate_models) - set(
        metadata_climate_models
    )
    if timeseries_climate_models_extras:
        error_msg = (
            "climate_model(s) `{}` in timeseries is not found in metadata".format(
                timeseries_climate_models_extras
            )
        )
        LOGGER.error(error_msg)
        climate_model_consistency = False

    # model_reported_climate_models_extras = set(model_reported_climate_models) - set(
    #     metadata_climate_models
    # )
    # if model_reported_climate_models_extras:
    #     error_msg = "climate_model(s) `{}` in model reported csv is not found in metadata".format(
    #         model_reported_climate_models_extras
    #     )
    #     LOGGER.error(error_msg)
    #     climate_model_consistency = False

    if comments is not None:
        comments_reported_climate_models_extras = set(metadata_climate_models) - set(
            comments_climate_models
        )
        if comments_reported_climate_models_extras:
            error_msg = "climate_model(s) `{}` in submission comments csv is not found in metadata".format(
                comments_reported_climate_models_extras
            )
            LOGGER.error(error_msg)
            climate_model_consistency = False

    if not climate_model_consistency:
        raise ValueError("climate_model not consistent across the bundle")

    LOGGER.info("valid submission bundle")

    return timeseries_consistent, metadata_consistent, comments_consistent


def validate_submission(submission, protocol=None):
    """
    Validate that an RCMIP submission complies with the required data format

    Parameters
    ----------
    submission : :obj:`scmdata.ScmRun`
        Data to validate

    protocol : str
        Data file containing the RCMIP protocol against which to validate the
        data. If ``None``, the submission template will be loaded from
        ``pyrcmip/data/rcmip-data-submission-template-v4-0-0.xlsx``.

    Returns
    -------
    :obj:`scmdata.ScmRun`
        Input data, converted to match RCMIP units

    Raises
    ------
    ProtocolConsistencyError
        The data is not consistent with the protocol
    """
    raise_error = False

    protocol_variables = load_submission_template_definitions(protocol, "variables")

    try:
        validate_variables(
            submission.get_unique_meta("variable"),
            protocol_variables["variable"].tolist(),
        )
        LOGGER.info("valid variables")
    except ProtocolConsistencyError:
        LOGGER.exception("problems with variables in submission")
        raise_error = True

    try:
        validate_regions(
            submission.get_unique_meta("region"),
            load_submission_template_definitions(protocol, "regions")[
                "region"
            ].tolist(),
        )
        LOGGER.info("valid regions")
    except ProtocolConsistencyError:
        LOGGER.exception("problems with regions in submission")
        raise_error = True

    try:
        validate_scenarios(
            submission.get_unique_meta("scenario"),
            load_submission_template_definitions(protocol, "scenarios")[
                "scenario_id"
            ].tolist(),
        )
        LOGGER.info("valid scenarios")
    except ProtocolConsistencyError:
        LOGGER.exception("problems with scenarios in submission")
        raise_error = True

    required_cols = ("climate_model", "ensemble_member")
    for required_col in required_cols:
        if required_col not in submission.meta:
            LOGGER.exception(
                "no {} metadata included in submission".format(required_col)
            )
            raise_error = True

    if "unit" not in submission.meta:
        LOGGER.exception("no unit metadata included in submission")
        raise_error = True
    else:
        try:
            out = convert_units_to_rcmip_units(submission, protocol_variables)
            LOGGER.info("succesfully converted units")
        except ProtocolConsistencyError:
            LOGGER.exception("problems converting units in submission")
            raise_error = True

    if raise_error:
        raise ProtocolConsistencyError(
            "submission is inconsistent with the RCMIP protocol"
        )

    LOGGER.info("valid submission")

    return out


def _validate_sets(to_check, protocol, error_str_id):
    not_included = set(to_check) - set(protocol)

    if not_included:
        raise ProtocolConsistencyError(
            "{} do not match the RCMIP protocol:\n{}".format(
                error_str_id, sorted(not_included)
            )
        )


def validate_variables(vars_to_check, protocol_variables):
    """
    Validate variables against variables in the RCMIP protocol

    Parameters
    ----------
    vars_to_check : list-like
        Variables to check

    protocol_variables : list-like
        Variables in the RCMIP protocol

    Raises
    ------
    ProtocolConsistencyError
        ``vars_to_check`` contains variables not included in ``protocol_variables``
    """
    _validate_sets(vars_to_check, protocol_variables, "Variables")


def validate_regions(regions_to_check, protocol_regions):
    """
    Validate regions against regions in the RCMIP protocol

    Parameters
    ----------
    regions_to_check : list-like
        Regions to check

    protocol_regions : list-like
        Regions in the RCMIP protocol

    Raises
    ------
    ProtocolConsistencyError
        ``regions_to_check`` contains regions not included in ``protocol_regions``
    """
    _validate_sets(regions_to_check, protocol_regions, "Regions")


def validate_scenarios(scenarios_to_check, protocol_scenarios):
    """
    Validate scenarios against scenarios in the RCMIP protocol

    Parameters
    ----------
    scenarios_to_check : list-like
        Scenarios to check

    protocol_scenarios : list-like
        Scenarios in the RCMIP protocol

    Raises
    ------
    ProtocolConsistencyError
        ``scenarios_to_check`` contains scenarios not included in ``protocol_scenarios``
    """
    _validate_sets(scenarios_to_check, protocol_scenarios, "scenarios")


def convert_units_to_rcmip_units(submission, protocol_variables):
    """
    Convert units to RCMIP units

    Parameters
    ----------
    submission : :obj:`scmdata.ScmRun`
        Submission to convert

    protocol_variables : :obj:`pd.DataFrame`
        Variables and units as defined by the RCMIP protocol

    Returns
    -------
    :obj:`scmdata.ScmRun`
        Submission with units converted to RCMIP units

    Raises
    ------
    ProtocolConsistencyError
        Units could not be converted to RCMIP units
    """
    res = []
    for vgroup in submission.groupby("variable"):
        res.append(_convert_to_rcmip_units(vgroup, protocol_variables))

    errors = [r for r in res if isinstance(r, dict)]
    if errors:
        errors = pd.DataFrame(errors)

        raise ProtocolConsistencyError(
            "Could not convert submission to protocol units. Errors:\n{}".format(errors)
        )

    return run_append(res)


def _convert_to_rcmip_units(variable_group, protocol_variables):
    variable = variable_group.get_unique_meta("variable", no_duplicates=True)

    if variable.startswith("Emissions|NOx"):
        conversion_context = "NOx_conversions"
    else:
        conversion_context = None

    try:
        protocol_unit = protocol_variables[protocol_variables["variable"] == variable][
            "unit"
        ].iloc[0]
        out = variable_group.convert_unit(protocol_unit, context=conversion_context)
        if conversion_context:
            out = out.drop_meta("unit_context", inplace=False)

        return out
    except IndexError as exc:
        input_unit = variable_group.get_unique_meta("unit", no_duplicates=True)
        return {
            "variable": variable,
            "input_unit": input_unit,
            "protocol_unit": "None",
            "exc_type": type(exc),
            "exc_message": str(exc),
        }
    except (pint.errors.DimensionalityError, pint.errors.UndefinedUnitError) as exc:
        input_unit = variable_group.get_unique_meta("unit", no_duplicates=True)
        return {
            "variable": variable,
            "input_unit": input_unit,
            "protocol_unit": protocol_unit,
            "exc_type": type(exc),
            "exc_message": str(exc),
        }


def _check_expected_columns(inp, expected_columns):
    correct_columns = set(inp.columns) == expected_columns
    if not correct_columns:
        raise ProtocolConsistencyError(
            "Input columns: {}. Expected columns: {}.".format(
                set(inp.columns), expected_columns
            )
        )


def validate_submission_model_reported_metrics(inp):
    """
    Validate a submission of model reported metrics

    Parameters
    ----------
    inp : :obj:`pd.DataFrame`
        Input to validate

    Returns
    -------
    :obj:`pd.DataFrame`
        Validated input

    Raises
    ------
    ProtocolConsistencyError
        The columns of res are not as expected (i.e.
        ``{"value", "ensemble_member", "RCMIP name", "unit", "climate_model"}``),
        more than one climate model is included in ``res``, the ``ensemble_member``
        column is not integers, an unrecognised metric is provided or the provided
        unit is not compatible with RCMIP.
    """
    out = inp.copy()

    expected_columns = {
        "value",
        "ensemble_member",
        "RCMIP name",
        "unit",
        "climate_model",
    }
    _check_expected_columns(out, expected_columns)

    if (out["ensemble_member"].astype(int) == out["ensemble_member"]).all():
        out["ensemble_member"] = out["ensemble_member"].astype(int)
    else:
        raise ProtocolConsistencyError("`ensemble_member` column must be integers")

    # TODO: better system for this
    allowed_metrics = ["Equilibrium Climate Sensitivity"]
    allowed_units = ["K"]
    if not all(out["RCMIP name"].isin(allowed_metrics)):
        raise ProtocolConsistencyError(
            "The `RCMIP name` column should only contain 'Equilibrium Climate Sensitivity"
        )

    if not all(out["unit"].isin(allowed_units)):
        raise ProtocolConsistencyError("values must be provided in 'K'")

    return out


def validate_submission_model_meta(inp):
    """
    Validate a submission's metadata

    Parameters
    ----------
    inp : :obj:`pd.DataFrame`
        Metadata submission to validate

    Returns
    -------
    :obj:`pd.DataFrame`
        Validated metadata submission

    Raises
    ------
    ProtocolConsistencyError
        The columns of res are not as expected (i.e.
        ``{"climate_model", "climate_model_name", "climate_model_version", "climate_model_configuration_label", "climate_model_configuration_description", "project", "name_of_person", "literature_reference"}``).
    """
    out = inp.copy()
    expected_columns = {
        "climate_model",
        "climate_model_name",
        "climate_model_version",
        "climate_model_configuration_label",
        "climate_model_configuration_description",
        "project",
        "name_of_person",
        "literature_reference",
    }
    _check_expected_columns(out, expected_columns)

    return out


def validate_submission_model_comments(inp):
    """
    Validate a submission's comments

    Parameters
    ----------
    inp : :obj:`pd.DataFrame`
        Submission comments to validate

    Returns
    -------
    :obj:`pd.DataFrame`
        Validated metadata submission

    Raises
    ------
    ProtocolConsistencyError
        The columns of res are not as expected (i.e.
        ``{"name","comment","email","climate_model","scenario","region","variable","year",}``).
    """
    out = inp.copy()
    expected_columns = {
        "name",
        "comment",
        "email",
        "climate_model",
        "scenario",
        "region",
        "variable",
        "year",
    }
    _check_expected_columns(out, expected_columns)

    return out



"""
Console script for rcmip.
"""

import gzip
import logging
import sys
from os.path import basename, join
from pathlib import Path

import boto3
import click
import jwt
import pandas as pd
import semver
import xarray as xr
from botocore.exceptions import ClientError

from pyrcmip.io import (
    _upload_file,
    read_results_submission,
    read_submission_model_comments,
    read_submission_model_metadata,
)
from pyrcmip.utils import tabular_to_netcdf
from pyrcmip.validate import validate_submission_bundle

LOGGER = logging.getLogger(__name__)
DEFAULT_LOG_FORMAT = "{process} {asctime} {levelname}:{name}:{message}"


class ColorFormatter(logging.Formatter):
    """
    Colour formatter for log messages

    A handy little tool for making our log messages look slightly prettier
    """

    colors = {
        "DEBUG": dict(fg="blue"),
        "INFO": dict(fg="green"),
        "WARNING": dict(fg="yellow"),
        "Error": dict(fg="red"),
        "ERROR": dict(fg="red"),
        "EXCEPTION": dict(fg="red"),
        "CRITICAL": dict(fg="red"),
    }

    def format(self, record):
        """
        Format a record so it has pretty colours

        Parameters
        ----------
        record : :obj:`logging.LogRecord`
            Record to format

        Returns
        -------
        str
            Formatted message string
        """
        formatted_message = super(ColorFormatter, self).format(record)

        level = record.levelname

        if level in self.colors:
            level_colour = click.style("{}".format(level), **self.colors[level])
            formatted_message = formatted_message.replace(level, level_colour)

        return formatted_message


class ClickHandler(logging.Handler):
    """
    Handler which emits using click when going to stdout
    """

    _use_stderr = True

    def emit(self, record):
        """
        Emit a record

        Parameters
        ----------
        record : :obj:`logging.LogRecord`
            Record to emit
        """
        try:
            msg = self.format(record)
            click.echo(msg, err=self._use_stderr)

        except Exception:  # pragma: no cover
            self.handleError(record)


_default_handler = ClickHandler()
_default_handler.formatter = ColorFormatter(DEFAULT_LOG_FORMAT, style="{")


@click.group(name="rcmip")
@click.option(
    "--log-level",
    default="INFO",
    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR", "EXCEPTION", "CRITICAL"]),
)
def cli(log_level):
    """
    Command-line interface for pyrcmip
    """
    root = logging.getLogger()
    root.handlers.append(_default_handler)
    root.setLevel(log_level)

    logging.captureWarnings(True)


def _read_bundle(timeseries, metadata, comments=None):
    try:
        scmrun = read_results_submission(timeseries)
    except Exception as e:
        LOGGER.exception("reading timeseries failed")
        raise click.ClickException(str(e))

    # Model metrics are not requested in the latest RCMIP round, but keeping it
    # around in case we want to re-enable it in the future.
    # try:
    #     model_reported_df = read_submission_model_reported(model_reported)
    # except Exception as e:
    #     LOGGER.exception("reading model_reported failed")
    #     raise click.ClickException(str(e))

    try:
        metadata_df = read_submission_model_metadata(metadata)
    except Exception as e:
        LOGGER.exception("reading metadata failed")
        raise click.ClickException(str(e))

    if comments is not None:
        try:
            comments_df = read_submission_model_comments(comments)
        except Exception as e:
            LOGGER.exception("reading comments failed")
            raise click.ClickException(str(e))
    else:
        comments_df = None

    return scmrun, metadata_df, comments_df


timeseries = click.argument(
    "timeseries", nargs=-1, required=True, type=click.Path(exists=True, dir_okay=False)
)
# Model metrics are not requested in the latest RCMIP round, but keeping it
# around in case we want to re-enable it in the future.
# model_reported = click.argument(
#     "model_reported",
#     nargs=1,
#     required=True,
#     type=click.Path(exists=True, dir_okay=False),
# )
metadata = click.argument(
    "metadata", nargs=1, required=True, type=click.Path(exists=True, dir_okay=False)
)


@cli.command()
@timeseries
@metadata
@click.option(
    "--comments",
    default=None,
    help="CSV with comments to upload along data.",
    type=click.Path(exists=True, dir_okay=False),
)
def validate(timeseries, metadata, comments=None):
    """
    Validate submission input.

    Two different types of input data are required for validation, namely:

    One or more ``TIMESERIES`` files in which the timeseries output is stored. These should be
    CSV or NetCDF files conforming to the format expected by ``scmdata``. Multiple
    timeseries inputs can be specified, but care must be taken to ensure that all of
    the individual timeseries have unique metadata.

    ``METADATA`` is the CSV file in which the metadata output is stored.

    Optionally, a CSV file with comments about the data to be uploaded can be included in the
    submission with the --comments flag. If included, it will also be validated.
    """
    for timeseries_fname in timeseries:
        scmrun, metadata_df, comments_df = _read_bundle(
            timeseries_fname, metadata, comments
        )

        try:
            validate_submission_bundle(scmrun, metadata_df, comments_df)
        except Exception as e:
            raise click.ClickException(str(e))


def validate_version(ctx, param, value):
    """
    Validate version string

    Parameters
    ----------
    ctx
        Not used

    param
        Not used

    value : str
        Version string to validate

    Returns
    -------
    str
        Validated version string

    Raises
    ------
    :obj:`click.BadParameter`
        Version string cannot be passed or does not follow semantic versioning
    """
    try:
        s = semver.VersionInfo.parse(value)

        if s.prerelease is None and s.build is None:
            return value
        else:
            raise click.BadParameter(
                "Version must only contain major, minor and patch values"
            )
    except ValueError:
        raise click.BadParameter("Cannot parse version string")


@cli.command()
@click.option(
    "--token",
    required=True,
    help="Authentication token. Contact eearp@leeds.ac.uk for a token",
)
@click.option("--bucket", default="alexrom-ns9188k-rcmip")
@click.option("--model", required=True)
@click.option(
    "--version",
    required=True,
    callback=validate_version,
    help="Version of the data being uploaded. Must be a valid semver version string (https://semver.org/). "
    "For example 2.0.0",
)
@timeseries
# @model_reported
@metadata
@click.option(
    "--comments",
    default=None,
    help="CSV with comments to upload along data.",
    type=click.Path(exists=True, dir_okay=False),
)
@click.option(
    "--endpoint", default="https://s3.nird.sigma2.no", help="S3 endpoint (for testing)"
)
def upload(token, bucket, model, version, timeseries, metadata, comments, endpoint):
    """
    Validate and upload data to RCMIP's S3 bucket.

    All the files for a given version have to be uploaded together. Notice that this command
    will convert the data into a netCDF file prior to uploading it to the remote server if not
    already in that format.

    One or more ``TIMESERIES`` files in which the timeseries output is stored. These should be
    CSV or NetCDF files conforming to the format expected by ``scmdata``. Multiple
    timeseries inputs can be specified, but care must be taken to ensure that all of
    the individual timeseries have unique metadata. Each timeseries file will be validated and
    uploaded independently. Each file must contain all the variables pertaining to any simulation
    scenarios included in that file.

    ``METADATA`` is the CSV file in which the metadata output is stored.

    Optionally, a CSV file with comments about the data to be uploaded can be included in the
    submission with the --comments flag.
    """
    # Prepare client to upload data to S3
    t = jwt.decode(token, options={"verify_signature": False})
    session = boto3.session.Session(
        aws_access_key_id=t["access_key_id"],
        aws_secret_access_key=t["secret_access_key"],
    )
    client = session.client("s3", endpoint_url=endpoint)

    root_key = "{}/{}/{}".format(t["org"], model, version)

    for file_idx, timeseries_fname in enumerate(timeseries):
        LOGGER.info("Reading and validating {}".format(timeseries_fname))
        try:
            scmrun, metadata_df, comments_df = _read_bundle(
                timeseries_fname, metadata, comments
            )
            (
                scmrun_rcmip_compatible,
                metadata_rcmip_compatible,
                comments_rcmip_compatible,
            ) = validate_submission_bundle(scmrun, metadata_df, comments_df)
        except Exception as e:
            LOGGER.exception(e)
            raise click.ClickException("Validation failed. Fix issues and rerun")

        # Check if this version is already uploaded (using the {key}-complete dummy file)
        try:
            LOGGER.debug("Checking if object with key {} exists".format(root_key))
            client.head_object(Bucket=bucket, Key=root_key + "-complete")

            raise click.ClickException(
                "Data for this version has already been uploaded. Increment the version and try again"
            )
        except ClientError:
            LOGGER.debug("Object with key {} does not exist".format(root_key))

        def _get_fname(data_type, scenario=None):
            if data_type == "data":
                if scenario:
                    fname = "rcmip-{}-{}-{}-{}-{:03}".format(
                        model, version, data_type, scenario, file_idx
                    )
                else:
                    fname = "rcmip-{}-{}-{}-{:03}".format(
                        model, version, data_type, file_idx
                    )
            else:
                fname = "rcmip-{}-{}-{}".format(model, version, data_type)
            return join(root_key, fname)

        # Create netcdf file with SCM data
        scenarios_to_upload = tabular_to_netcdf(
            scmrun_rcmip_compatible, metadata_rcmip_compatible, version=version
        )

        for scenario, df in scenarios_to_upload.items():
            # upload it
            _upload_file(
                client,
                bucket=bucket,
                key=_get_fname("data", scenario),
                run=df,
                file_format="nc",
            )

        # If any comments were detected, upload them to the folder
        if comments is not None:
            _upload_file(
                client,
                bucket=bucket,
                key=_get_fname("comments"),
                run=comments_rcmip_compatible,
                file_format="csv",
            )

    # Finally mark the upload as complete by uploading a dummy file
    # Writing this dummy file will be used to start the processing of the upload
    client.put_object(Bucket=bucket, Key=root_key + "-complete")

    LOGGER.info("All files uploaded successfully")


@cli.command()
@click.option(
    "--token",
    required=True,
    help="Authentication token. Contact eearp@leeds.ac.uk for a token",
)
@click.option("--bucket", default="alexrom-ns9188k-rcmip")
@click.option("--model", required=True)
@click.option(
    "--version",
    required=True,
    callback=validate_version,
    help="Version of the data that was uploaded. Must be a valid semver version string (https://semver.org/). "
    "For example 2.0.0",
)
@click.option(
    "--unzip",
    is_flag=True,
    default=False,
    help="If set, unzip .gz files directly instead of saving compressed file.",
)
@click.argument("outdir", required=True, type=click.Path(exists=True, dir_okay=True))
@click.option(
    "--endpoint", default="https://s3.nird.sigma2.no", help="S3 endpoint (for testing)"
)
def download(token, bucket, model, version, unzip, outdir, endpoint):
    """
    Download submitted files.
    """
    t = jwt.decode(token, options={"verify_signature": False})
    session = boto3.session.Session(
        aws_access_key_id=t["access_key_id"],
        aws_secret_access_key=t["secret_access_key"],
    )
    client = session.client("s3", endpoint_url=endpoint)

    root_key = "{}/{}/{}/".format(t["org"], model, version)  # trailing / is important

    resp = client.list_objects(Bucket=bucket, Prefix=root_key)

    if "Contents" not in resp:
        raise click.ClickException(
            "No files for {}=={} have been uploaded".format(model, version)
        )

    for o in resp["Contents"]:
        key = o["Key"]
        LOGGER.info("Downloading {}".format(key))
        resp = client.get_object(Bucket=bucket, Key=key)

        local_path = join(outdir, basename(key))

        if unzip and local_path.endswith(".gz"):
            local_path = local_path[:-3]  # remove .gz from saved file name
            LOGGER.info("Unzipping {} -> {}".format(key, local_path))

            # Stream gzip from S3 and write uncompressed to disk
            with gzip.open(resp["Body"], "rb") as f_in, open(local_path, "wb") as f_out:
                for chunk in iter(lambda: f_in.read(1024 * 1024), b""):
                    f_out.write(chunk)
        else:
            # Save as-is
            with open(local_path, "wb") as f:
                for chunk in resp["Body"]:
                    f.write(chunk)

    LOGGER.info("All files downloaded successfully")


@cli.command()
@click.argument(
    "file", nargs=1, required=True, type=click.Path(exists=True, dir_okay=False)
)
@click.argument("outdir", required=True, type=click.Path(exists=True, dir_okay=True))
def convert_to_csvs(file, outdir):
    """
    Convert a NetCDF scenario file to CSVs.

    This command will produce two CSV files:

    - a main data CSV file with the numerical values of the scenario variables

    - a metadata CSV file
    """
    LOGGER.info("Loading {}".format(file))
    ds = xr.open_dataset(file, engine="netcdf4")

    # Extract attributes
    scenario = ds.attrs.get("scenario", "unknown")
    model_name = ds.attrs.get("climate_model", "unknown")
    version = ds.attrs.get("version", "unknown")

    # Build metadata CSV
    # Grab only keys that are in our mapping
    metadata_keys = [k for k in ds.attrs.keys()]
    metadata_values = [ds.attrs.get(k) for k in metadata_keys]

    metadata_df = pd.DataFrame([metadata_keys, metadata_values])
    metadata_output = (
        Path(outdir) / f"rcmip-{model_name}-{version}-metadata-{scenario}.csv"
    )
    metadata_df.to_csv(metadata_output, index=False, header=False)

    LOGGER.info("Successfully created metadata CSV file: {}".format(metadata_output))

    # Build data CSV
    rows = []
    time = ds.coords["time"].values
    for var in ds.data_vars:
        unit = ds[var].attrs.get("unit", "unknown")
        for em in ds.coords["ensemble_member"].values:
            for re in ds.coords["region"].values:
                values = ds[var].sel(ensemble_member=em, region=re).values
                row = {
                    "climate_model": model_name,
                    "model": ds.attrs.get("model", "unknown"),
                    "scenario": scenario,
                    "variable": var,
                    "ensemble_member": em,
                    "region": re,
                    "unit": unit,
                }
                row.update({str(t): v for t, v in zip(time, values)})
                rows.append(row)

    df_out = pd.DataFrame(rows)
    data_output = Path(outdir) / f"rcmip-{model_name}-{version}-data-{scenario}.csv"
    df_out.to_csv(data_output, index=False)

    LOGGER.info("Successfully created values CSV file: {}".format(data_output))


def run_cli():
    """
    Run command-line interface

    TODO: fix this so environment variables can be used
    """
    sys.exit(cli(auto_envvar_prefix="RCMIP"))  # pragma: no cover


if __name__ == "__main__":
    run_cli()  # pragma: no cover
